from asn1crypto import pem
from asn1crypto.x509 import Certificate


def _load_certificate(cert_bytes: bytes) -> Certificate:
    if pem.detect(cert_bytes):
        _, _, cert_bytes = pem.unarmor(cert_bytes)

    return Certificate.load(cert_bytes)



from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey

from pyattest.configs.config import Config
from pyattest.exceptions import PyAttestException


class Assertion:
    def __init__(
        self,
        raw: bytes,
        expected_hash: bytes,
        public_key: EllipticCurvePublicKey,
        config: Config,
    ):
        self.raw = raw
        self.expected_hash = expected_hash
        self.public_key = public_key
        self.config = config

        self.verified = False
        self._verified_data = {}

    def verify(self):
        self.config.assertion_verifier_class(self).verify()

    @property
    def data(self):
        if not self.verified:
            raise PyAttestException(
                "Assertion needs to be verified before accessing it."
            )

        return self._verified_data

    def verified_data(self, data: dict):
        """The verifier from the config can set all relevant data once the verification is complete."""
        self.verified = True
        self._verified_data = data



import os
from hashlib import sha256
from pathlib import Path

import pytest
from cryptography.hazmat.primitives import serialization
from cryptography.x509.base import load_pem_x509_certificate

from pyattest.assertion import Assertion
from pyattest.configs.apple import AppleConfig
from pyattest.testutils.factories.attestation import apple as apple_factory

root_ca = load_pem_x509_certificate(
    Path("pyattest/testutils/fixtures/root_cert.pem").read_bytes()
)
root_ca_pem = root_ca.public_bytes(serialization.Encoding.PEM)
nonce = os.urandom(32)


@pytest.mark.skip(reason="not yet fully implemented")
def test_happy_path():
    """Test the basic attest verification where everything works like it should :)"""
    attest, public_key = apple_factory.get(app_id="foo", nonce=nonce)
    key_id = sha256(public_key).digest()
    config = AppleConfig(
        key_id=key_id, app_id="foo", root_ca=root_ca_pem, production=False
    )

    raw_assertion = b"asdf"
    expected_hash = b"asdf"

    assertion = Assertion(raw_assertion, expected_hash, public_key, config)
    assertion.verify()



import datetime

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.x509.oid import NameOID

key_usage = x509.KeyUsage(
    True,  # digital_signature
    False,  # content_commitment
    False,  # key_encipherment
    False,  # data_encipherment
    False,  # key_agreement
    True,  # key_cert_sign
    True,  # crl_sign
    False,  # encipher_only
    False,  # decipher_only
)


def generate():
    """Generate root and intermediate certificates for the apple attestation."""
    root_private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    with open("pyattest/testutils/fixtures/root_key.pem", "wb") as f:
        f.write(
            root_private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.BestAvailableEncryption(b"123"),
            )
        )

    subject = issuer = x509.Name(
        [x509.NameAttribute(NameOID.ORGANIZATION_NAME, "pyattest-testing-ca")]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(root_private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=10))
        .add_extension(key_usage, critical=False)
        .sign(root_private_key, hashes.SHA256())
    )

    with open("pyatest/testutils/fixtures/root_cert.pem", "wb") as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))



import os
from hashlib import sha256
from pathlib import Path

import pytest
from cryptography.hazmat.primitives import serialization
from cryptography.x509.base import load_pem_x509_certificate
from pytest import raises

from pyattest.attestation import Attestation
from pyattest.configs.apple import AppleConfig
from pyattest.exceptions import (
    InvalidAaguidException,
    InvalidNonceException,
    InvalidKeyIdException,
    InvalidAppIdException,
    InvalidCounterException,
    InvalidCredentialIdException,
)
from pyattest.testutils.factories.attestation import apple as apple_factory

root_ca = load_pem_x509_certificate(
    Path("pyattest/testutils/fixtures/root_cert.pem").read_bytes()
)
root_ca_pem = root_ca.public_bytes(serialization.Encoding.PEM)
nonce = os.urandom(32)


@pytest.mark.asyncio
async def test_happy_path():
    """Test the basic attest verification where everything works like it should :)"""
    attest, public_key = apple_factory.get(app_id="foo", nonce=nonce)
    key_id = sha256(public_key).digest()
    config = AppleConfig(
        key_id=key_id, app_id="foo", root_ca=root_ca_pem, production=False
    )

    attestation = Attestation(attest, nonce, config)
    await attestation.verify()


@pytest.mark.asyncio
async def test_aaguid():
    attest, public_key = apple_factory.get(app_id="foo", nonce=nonce)
    key_id = sha256(public_key).digest()
    config = AppleConfig(
        key_id=key_id, app_id="foo", root_ca=root_ca_pem, production=True
    )

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidAaguidException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_nonce():
    attest, public_key = apple_factory.get(app_id="foo", nonce=nonce)
    key_id = sha256(public_key).digest()
    config = AppleConfig(
        key_id=key_id, app_id="foo", root_ca=root_ca_pem, production=False
    )

    attestation = Attestation(attest, os.urandom(32), config)
    with raises(InvalidNonceException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_key_id():
    attest, public_key = apple_factory.get(app_id="foo", nonce=nonce)
    key_id = sha256(b"invalid_public_key").digest()
    config = AppleConfig(
        key_id=key_id, app_id="foo", root_ca=root_ca_pem, production=False
    )

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidKeyIdException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_app_id():
    attest, public_key = apple_factory.get(app_id="foo", nonce=nonce)
    key_id = sha256(public_key).digest()
    config = AppleConfig(
        key_id=key_id, app_id="bar", root_ca=root_ca_pem, production=False
    )

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidAppIdException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_counter():
    attest, public_key = apple_factory.get(
        app_id="foo", nonce=nonce, aaguid=b"appattestdevelop", counter=9
    )
    key_id = sha256(public_key).digest()
    config = AppleConfig(
        key_id=key_id, app_id="foo", root_ca=root_ca_pem, production=False
    )

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidCounterException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_credential_id():
    attest, public_key = apple_factory.get(
        app_id="foo", nonce=nonce, wrong_public_key=True
    )
    key_id = sha256(public_key).digest()
    config = AppleConfig(
        key_id=key_id, app_id="foo", root_ca=root_ca_pem, production=False
    )

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidCredentialIdException):
        await attestation.verify()



import base64
from pathlib import Path

import pytest
from asn1crypto import pem, x509

from pyattest.attestation import Attestation
from pyattest.configs.apple import AppleConfig
from pyattest.configs.google import GoogleConfig


@pytest.mark.skip(reason="only internal")
def test_apple():
    """Special test specific for the attest_apple package."""
    server_nonce = "1fc6d08a2ffc842e25e4ca0deac203cd"  # sent beforehand
    key_id = "mbrDsK6QyPjKoTiliNSETympZqA643NiWIiK6B7vEOw="  # SHA-256 uf the public key which is in the request
    uid = "registration;A6215681-970F-4761-B352-0F0735F7E86F"  # in request header

    config = AppleConfig(
        key_id=base64.b64decode(key_id),
        app_id="5LVDC4HW22.ch.dreipol.dreiAttestTestHost",
        production=False,
    )
    attest = Path("tests/fixtures/attest_apple").read_text().rstrip()

    # dreiAttest specific way of generating the attestation nonce
    nonce = (uid + key_id + server_nonce).encode()

    attestation = Attestation(base64.b64decode(attest), nonce, config)
    attestation.verify()


@pytest.mark.skip(reason="only internal")
def test_google():
    nonce = str.encode("f81d4fae-7dec-11d0-a765-00a0c91e6bf6")
    cert = """-----BEGIN CERTIFICATE-----
MIIC4jCCAcoCAQEwDQYJKoZIhvcNAQEFBQAwNzEWMBQGA1UEAwwNQW5kcm9pZCBE
ZWJ1ZzEQMA4GA1UECgwHQW5kcm9pZDELMAkGA1UEBhMCVVMwHhcNMTkxMTExMTMw
NzAxWhcNNDkxMTAzMTMwNzAxWjA3MRYwFAYDVQQDDA1BbmRyb2lkIERlYnVnMRAw
DgYDVQQKDAdBbmRyb2lkMQswCQYDVQQGEwJVUzCCASIwDQYJKoZIhvcNAQEBBQAD
ggEPADCCAQoCggEBAMPuob6rd30zvriza1BUUQEw2k8bEcc5BF/L+H25XdXw3boQ
niM5VFrsThnad//Pf/piOT6Rhd2Vj6DlnY8jCS2aTLOsFu3DPqTI/wcYVIM6iYqw
8X4eW20vbzP8lHYCq+b5PSa1ikaQocsl/4gi4JC9KvCW1M/EJpLPV6SsHoDQLkHf
ahQ7YEL/4pzXzfFW4B+nILB4ez2tjPEv17d2G7LnF14DIqejYVEutjVzT0wjUzCL
w6ZGYeKqcPI9WEsw+e5JXHh37CN4GtuaARcUkzYnqgGCuQqGapV3JHSLh7Zh070W
lNomd6mgFCOylkpwCP1Gs+0MuVBmS0RGOazAhi8CAwEAATANBgkqhkiG9w0BAQUF
AAOCAQEAF2fbDSxygW7ig5VA+a2W7s2PrIpA1Ny6Xh7O0nceTWMnypO5H1bv++M2
/PgSYX5YdCbgXg6rus3WL7PGl40SMwyZde9yTyB8n4DXEzL3faa7SQSPdWpolBFV
SAYHc3ViOTTT1QtHrRj47VKCPGs873/hDM6VSfeonPN6HiS3b70h4c4Mjt8/Aq9g
gmrBM2lV/sP0yrlkA73bEXMMo1HWC/MD8QTc1cREy/3C91iQsIoqY4NJ4ao4bwOf
269SioU5Eko9knyJfDBtX2EDdOfYvy2jbH8Lm7G0v1TJlVXPXnvoQpukRfoclYRF
/ztOeYT8xl4DpIGzVYGarxLnYXK0qA==
-----END CERTIFICATE-----"""
    type_name, headers, der_bytes = pem.unarmor(cert.encode())
    cert = x509.Certificate.load(der_bytes)

    config = GoogleConfig(
        key_ids=[base64.b64encode(cert.sha256)],
        apk_package_name="ch.dreipol.rezhycle",
        production=False,
    )
    attest = Path("tests/fixtures/attest_google").read_text().rstrip()

    attestation = Attestation(attest, nonce, config)
    attestation.verify()



import base64
from pathlib import Path

import pytest
from cryptography.hazmat.primitives import serialization
from cryptography.x509.base import load_pem_x509_certificate
from pytest import raises

from pyattest.attestation import Attestation
from pyattest.configs.google import GoogleConfig
from pyattest.exceptions import (
    InvalidKeyIdException,
    InvalidCtsProfile,
    InvalidBasicIntegrity,
    InvalidCertificateChainException,
    InvalidAppIdException,
)
import os

from pyattest.testutils.factories.attestation import google as google_factory

root_cn = "pyattest-testing-leaf.ch"
root_ca = load_pem_x509_certificate(
    Path("pyattest/testutils/fixtures/root_cert.pem").read_bytes()
)
root_ca_pem = root_ca.public_bytes(serialization.Encoding.PEM)
nonce = os.urandom(32)


@pytest.mark.asyncio
async def test_happy_path():
    """Test the basic attest verification where everything works like it should :)"""
    attest, key_id = google_factory.get(apk_package_name="foo", nonce=nonce)
    config = GoogleConfig(
        key_ids=[base64.b64encode(key_id)],
        apk_package_name="foo",
        root_cn=root_cn,
        root_ca=root_ca_pem,
        production=False,
    )

    attestation = Attestation(attest, nonce, config)
    await attestation.verify()


@pytest.mark.asyncio
async def test_key_id():
    attest, key_id = google_factory.get(apk_package_name="foo", nonce=nonce)
    config = GoogleConfig(
        key_ids=[base64.b64encode(b"100%wrong")],
        apk_package_name="foo",
        root_cn=root_cn,
        root_ca=root_ca_pem,
        production=False,
    )

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidKeyIdException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_cts_profile():
    attest, key_id = google_factory.get(
        apk_package_name="foo", nonce=nonce, cts_profile=False
    )
    config = GoogleConfig(
        key_ids=[base64.b64encode(key_id)],
        apk_package_name="foo",
        root_cn=root_cn,
        root_ca=root_ca_pem,
        production=True,
    )

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidCtsProfile):
        await attestation.verify()


@pytest.mark.asyncio
async def test_basic_integrity():
    attest, key_id = google_factory.get(
        apk_package_name="foo", nonce=nonce, basic_integrity=False
    )
    config = GoogleConfig(
        key_ids=[base64.b64encode(key_id)],
        apk_package_name="foo",
        root_cn=root_cn,
        root_ca=root_ca_pem,
        production=True,
    )

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidBasicIntegrity):
        await attestation.verify()


@pytest.mark.asyncio
async def test_apk_package_name():
    attest, key_id = google_factory.get(apk_package_name="foo", nonce=nonce)
    config = GoogleConfig(
        key_ids=[base64.b64encode(key_id)],
        apk_package_name="bar",
        root_cn=root_cn,
        root_ca=root_ca_pem,
        production=True,
    )

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidAppIdException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_certificate_chain():
    attest, key_id = google_factory.get(apk_package_name="foo", nonce=nonce)
    config = GoogleConfig(
        key_ids=[base64.b64encode(key_id)],
        apk_package_name="bar",
        root_cn=root_cn,
        production=True,
    )

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidCertificateChainException):
        await attestation.verify()



import base64
import os

import pytest
from pytest import raises
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from pyattest.attestation import Attestation
from pyattest.configs.google_play_integrity_api import GooglePlayIntegrityApiConfig
from pyattest.exceptions import (
    PyAttestException,
    InvalidNonceException,
    InvalidAppIdException,
    InvalidAppIntegrity,
    InvalidDeviceIntegrity,
    InvalidKeyIdException,
    IllegalConfigurationException,
)
from pyattest.testutils.factories.attestation.google import (
    get_play_integrity_api_attestation,
)

nonce = os.urandom(32)


@pytest.mark.asyncio
async def test_happy_path():
    apk_package_name = "foo"
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name, nonce, "PLAY_RECOGNIZED", ["MEETS_DEVICE_INTEGRITY"]
    )
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
    )

    attestation = Attestation(attestation, nonce, config)
    await attestation.verify()


@pytest.mark.asyncio
async def test_happy_path_sideloaded():
    apk_package_name = "foo"
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name, nonce, "UNRECOGNIZED_VERSION", ["MEETS_DEVICE_INTEGRITY"]
    )
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
        allow_non_play_distribution=True,
        verify_code_signature_hex=[
            "00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00"
        ],
    )

    attestation = Attestation(attestation, nonce, config)
    await attestation.verify()


@pytest.mark.asyncio
async def test_invalid_signed_attest_token():
    apk_package_name = "foo"
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name, nonce, "PLAY_RECOGNIZED", ["MEETS_DEVICE_INTEGRITY"]
    )
    wrong_aes_key = os.urandom(32)
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(wrong_aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
    )

    attestation = Attestation(attestation, nonce, config)
    with raises(PyAttestException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_invalid_nonce():
    apk_package_name = "foo"
    invalid_nonce = os.urandom(32)
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name, invalid_nonce, "PLAY_RECOGNIZED", ["MEETS_DEVICE_INTEGRITY"]
    )
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
    )
    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidNonceException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_invalid_apk_package():
    apk_package_name = "foo"
    wrong_apk_package_name = "foo_"
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        wrong_apk_package_name, nonce, "PLAY_RECOGNIZED", ["MEETS_DEVICE_INTEGRITY"]
    )
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
    )
    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidAppIdException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_invalid_app_integrity():
    apk_package_name = "foo"
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name, nonce, "PLAY_UNRECOGNIZED", ["MEETS_DEVICE_INTEGRITY"]
    )
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
    )
    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidAppIntegrity):
        await attestation.verify()


@pytest.mark.asyncio
async def test_invalid_key_sideloaded():
    apk_package_name = "foo"
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name, nonce, "UNRECOGNIZED_VERSION", ["MEETS_DEVICE_INTEGRITY"]
    )
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
        allow_non_play_distribution=True,
        verify_code_signature_hex=[
            "A0:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00"
        ],
    )

    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidKeyIdException):
        await attestation.verify()


@pytest.mark.asyncio
async def test_invalid_device_integrity():
    apk_package_name = "foo"
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name, nonce, "PLAY_RECOGNIZED", []
    )
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
    )
    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidDeviceIntegrity):
        await attestation.verify()


@pytest.mark.asyncio
async def test_strong_device_integrity():
    apk_package_name = "foo"
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name,
        nonce,
        "PLAY_RECOGNIZED",
        ["MEETS_DEVICE_INTEGRITY", "MEETS_BASIC_INTEGRITY", "MEETS_STRONG_INTEGRITY"],
    )
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
        required_device_verdict="MEETS_STRONG_INTEGRITY",
    )
    attestation = Attestation(attestation, nonce, config)
    await attestation.verify()


@pytest.mark.asyncio
async def test_missing_strong_device_integrity():
    apk_package_name = "foo"
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name,
        nonce,
        "PLAY_RECOGNIZED",
        ["MEETS_DEVICE_INTEGRITY", "MEETS_BASIC_INTEGRITY"],
    )
    config = GooglePlayIntegrityApiConfig(
        decryption_key=base64.standard_b64encode(aes_key).decode(),
        verification_key=base64.standard_b64encode(
            verification_key.public_bytes(
                encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
            )
        ).decode(),
        apk_package_name=apk_package_name,
        production=True,
        required_device_verdict="MEETS_STRONG_INTEGRITY",
    )
    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidDeviceIntegrity):
        await attestation.verify()


@pytest.mark.asyncio
def test_illegal_config1():
    apk_package_name = "foo"
    _, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name, nonce, "UNRECOGNIZED_VERSION", ["MEETS_DEVICE_INTEGRITY"]
    )
    with raises(IllegalConfigurationException):
        GooglePlayIntegrityApiConfig(
            decryption_key=base64.standard_b64encode(aes_key).decode(),
            verification_key=base64.standard_b64encode(
                verification_key.public_bytes(
                    encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
                )
            ).decode(),
            apk_package_name=apk_package_name,
            production=True,
            allow_non_play_distribution=True,
            verify_code_signature_hex=None,
        )


@pytest.mark.asyncio
def test_illegal_config2():
    apk_package_name = "foo"
    _, verification_key, aes_key = get_play_integrity_api_attestation(
        apk_package_name, nonce, "UNRECOGNIZED_VERSION", ["MEETS_DEVICE_INTEGRITY"]
    )
    with raises(IllegalConfigurationException):
        GooglePlayIntegrityApiConfig(
            decryption_key=base64.standard_b64encode(aes_key).decode(),
            verification_key=base64.standard_b64encode(
                verification_key.public_bytes(
                    encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo
                )
            ).decode(),
            apk_package_name=apk_package_name,
            production=True,
            allow_non_play_distribution=True,
            verify_code_signature_hex=[],
        )



from pathlib import Path

from asn1crypto.x509 import Certificate

from pyattest.configs.config import Config
from pyattest.verifiers.apple_assertion import AppleAssertionVerifier
from pyattest.verifiers.apple_attestation import AppleAttestationVerifier
from pyattest.verifiers.utils import _load_certificate


class AppleConfig(Config):
    attestation_verifier_class = AppleAttestationVerifier
    assertion_verifier_class = AppleAssertionVerifier

    def __init__(
        self, key_id: bytes, app_id: str, production: bool, root_ca: bytes = None
    ):
        self.key_id = key_id
        self.app_id = app_id
        self.production = production
        self._custom_root_ca = root_ca

    @property
    def oid(self) -> str:
        """Object Identifier of the certificate extension containing our nonce."""
        return "1.2.840.113635.100.8.2"

    @property
    def root_ca(self) -> Certificate:
        """
        Apples App Attestation Root CA. This can be overwritten for easier testing.

        See also: https://www.apple.com/certificateauthority/private/
        """
        if self._custom_root_ca:
            bytes = self._custom_root_ca
        else:
            folder = Path(__file__).parent / "../certificates"
            bytes = Path(folder / "Apple_App_Attestation_Root_CA.pem").read_bytes()

        return _load_certificate(bytes)



from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA

from pyattest.verifiers.assertion import AssertionVerifier


class GoogleAssertionVerifier(AssertionVerifier):
    def verify(self):
        self.assertion.public_key.verify(
            self.assertion.raw, self.assertion.expected_hash, ECDSA(hashes.SHA256())
        )



import base64
import datetime
import os
from pathlib import Path

import jwt
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa, ec
from cryptography.hazmat.primitives.serialization.base import load_pem_private_key
from cryptography.x509.base import load_pem_x509_certificate
from cryptography.x509.oid import NameOID
from jose import jwe, jws
from jose.constants import ALGORITHMS

from pyattest.testutils.factories.certificates import key_usage


def get_play_integrity_api_attestation(
    apk_package_name: str,
    nonce: bytes,
    app_integrity_verdict: str,
    device_integrity_verdict: [str],
):
    """Helper to create a fake google attestation."""
    root_key = ec.generate_private_key(ec.SECP256K1())
    aes_key = os.urandom(32)
    payload = {
        "requestDetails": {
            "nonce": base64.urlsafe_b64encode(nonce).decode(),
            "requestPackageName": apk_package_name,
        },
        "appIntegrity": {
            "appRecognitionVerdict": app_integrity_verdict,
            "packageName": apk_package_name,
            "certificateSha256Digest": ["AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"],
        },
        "deviceIntegrity": {
            "deviceRecognitionVerdict": device_integrity_verdict,
        },
    }
    jws_token = jws.sign(payload, root_key, algorithm=ALGORITHMS.ES256)
    return (
        jwe.encrypt(
            jws_token,
            aes_key,
            algorithm=ALGORITHMS.A256KW,
            encryption=ALGORITHMS.A256GCM,
        ),
        root_key.public_key(),
        aes_key,
    )


def get(
    apk_package_name: str,
    nonce: bytes,
    basic_integrity: bool = True,
    cts_profile: bool = True,
    apk_cert_digest: bytes = None,
):
    """Helper to create a fake google attestation."""
    here = os.path.abspath(os.path.dirname(__file__))
    fixtures = os.path.join(here, "..", "..", "fixtures")

    root_key = load_pem_private_key(
        Path(f"{fixtures}/root_key.pem").read_bytes(), b"123"
    )
    root_cert = load_pem_x509_certificate(
        Path(f"{fixtures}/root_cert.pem").read_bytes()
    )

    apk_cert_digest = apk_cert_digest or b"foobar"

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    subject = x509.Name(
        [x509.NameAttribute(NameOID.ORGANIZATION_NAME, "pyattest-testing-leaf")]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(root_cert.subject)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=10))
        .add_extension(key_usage, critical=False)
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName("pyattest-testing-leaf.ch")]),
            critical=False,
        )
        .sign(root_key, hashes.SHA256())
    )

    data = {
        "timestampMs": 9860437986543,
        "nonce": base64.b64encode(nonce).decode(),
        "apkPackageName": apk_package_name,
        "apkCertificateDigestSha256": [base64.b64encode(apk_cert_digest).decode()],
        "ctsProfileMatch": cts_profile,
        "basicIntegrity": basic_integrity,
        "evaluationType": "BASIC",
    }

    headers = {
        "x5c": [
            cert.public_bytes(serialization.Encoding.PEM)
            .decode()
            .replace("-----BEGIN CERTIFICATE-----\n", "")
            .replace("\n-----END CERTIFICATE-----\n", ""),
            root_cert.public_bytes(serialization.Encoding.PEM)
            .decode()
            .replace("-----BEGIN CERTIFICATE-----\n", "")
            .replace("\n-----END CERTIFICATE-----\n", ""),
        ]
    }

    return (
        jwt.encode(data, private_key, algorithm="RS256", headers=headers),
        apk_cert_digest,
    )



from typing import Optional, List

from asn1crypto.x509 import Certificate

from pyattest.configs.config import Config
from pyattest.verifiers.google_assertion import GoogleAssertionVerifier
from pyattest.verifiers.google_attestation import GoogleAttestationVerifier
from pyattest.verifiers.utils import _load_certificate


class GoogleConfig(Config):
    attestation_verifier_class = GoogleAttestationVerifier
    assertion_verifier_class = GoogleAssertionVerifier

    def __init__(
        self,
        key_ids: List[bytes],
        apk_package_name: str,
        production: bool,
        root_cn: str = "attest.android.com",
        root_ca: Optional[bytes] = None,
    ):
        self.key_ids = key_ids
        self.apk_package_name = apk_package_name
        self.production = production
        self.root_cn = root_cn
        self._custom_root_ca = _load_certificate(root_ca) if root_ca else None

    @property
    def root_ca(self) -> Optional[Certificate]:
        """This is only used for simplified unit testing."""
        return self._custom_root_ca



import struct
from hashlib import sha256

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA

from pyattest.verifiers.assertion import AssertionVerifier
from cbor2 import loads as cbor_decode


class AppleAssertionVerifier(AssertionVerifier):
    def verify(self):
        """
        Verify the assertion object accompanied with the request. Each verified assertion reestablishes the legitimacy
        of the client. You typically require this for requests that access sensitive or premium content.
        """
        unpacked = self.unpack(self.assertion.raw)
        authenticator_data = unpacked["authenticator_data"]
        nonce = sha256(authenticator_data + self.assertion.expected_hash).digest()

        self.assertion.public_key.verify(
            unpacked["raw"]["signature"], nonce, ECDSA(hashes.SHA256())
        )

    @staticmethod
    def unpack(raw: bytes) -> dict:
        """Extract in `verify` method mentioned relevant data from cbor encoded raw bytes input."""
        raw = cbor_decode(raw)

        return {
            "raw": raw,
            "authenticator_data": raw["authenticatorData"],
            "rp_id": raw["authenticatorData"][:32],
            "counter": struct.unpack("!I", raw["authenticatorData"][33:37])[0],
        }



import base64
import json
from typing import Optional

from jose import jwe, jws

from pyattest.exceptions import (
    PyAttestException,
    InvalidNonceException,
    InvalidAppIdException,
    InvalidAppIntegrity,
    InvalidKeyIdException,
    InvalidDeviceIntegrity,
)
from pyattest.verifiers.attestation import AttestationVerifier


class GooglePlayIntegrityAttestationVerifier(AttestationVerifier):
    async def verify(self):
        """
        Verify the given attestation based on the Google documentation. The attestation is a nested JWS in JWE object.

        Raises a PyAttestException as soon as one of the verification steps fails.

        See also:
        Google Documentation https://developer.android.com/google/play/integrity/verdict#decrypt-verify
        """
        payload = self.unpack(self.attestation.raw)
        self.check_request_details(payload.get("requestDetails"))
        if self.attestation.config.production:
            self.check_app_integrity(payload.get("appIntegrity"))
            self.check_device_integrity(payload.get("deviceIntegrity"))

        self.attestation.verified_data({"data": payload})

    def unpack(self, jwt_object: str) -> dict:
        try:
            jwe_token = jwt_object
            jws_token = jwe.decrypt(jwe_token, self.attestation.config.decryption_key)
            payload = jws.verify(
                jws_token, self.attestation.config.verification_key, "ES256"
            )
            return json.loads(payload)
        except Exception as exception:
            raise PyAttestException from exception

    def check_app_integrity(self, app_integrity):
        self.check_verdict(app_integrity.get("appRecognitionVerdict"))
        self.verify_signing_keys(app_integrity.get("certificateSha256Digest"))
        self.verify_apk_package_name(app_integrity.get("packageName"))

    def check_device_integrity(self, device_integrity):
        if self.attestation.config.required_device_verdict not in device_integrity.get(
            "deviceRecognitionVerdict"
        ):
            raise InvalidDeviceIntegrity

    def check_request_details(self, request_details):
        self.verify_nonce(request_details.get("nonce"))
        self.verify_apk_package_name(request_details.get("requestPackageName"))

    def verify_nonce(self, nonce: str):
        if base64.urlsafe_b64decode(nonce) != self.attestation.nonce:
            raise InvalidNonceException

    def check_verdict(self, verdict):
        if verdict == "PLAY_RECOGNIZED":
            pass
        elif (
            self.attestation.config.allow_non_play_distribution
            and verdict == "UNRECOGNIZED_VERSION"
        ):
            pass
        else:
            raise InvalidAppIntegrity

    def verify_signing_keys(self, digest):
        if not self.attestation.config.verify_code_signature:
            return

        if digest != self.attestation.config.verify_code_signature:
            raise InvalidKeyIdException

    def verify_apk_package_name(self, package_name: Optional[str]):
        if not package_name or package_name != self.attestation.config.apk_package_name:
            raise InvalidAppIdException



import datetime
import os
import struct
from hashlib import sha256
from pathlib import Path

from _cbor2 import dumps as cbor_encode
from asn1crypto.core import OctetString
from cryptography import x509
from cryptography.hazmat._oid import ObjectIdentifier
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization.base import load_pem_private_key
from cryptography.x509.base import load_pem_x509_certificate
from cryptography.x509.extensions import UnrecognizedExtension
from cryptography.x509.oid import NameOID

from pyattest.testutils.factories.certificates import key_usage


def get(
    app_id: str,
    nonce: bytes,
    aaguid: bytes = b"appattestdevelop",
    counter: int = 0,
    wrong_public_key: bool = False,
):
    """Helper to create a fake apple attestation."""
    here = os.path.abspath(os.path.dirname(__file__))
    fixtures = os.path.join(here, "..", "..", "fixtures")

    root_key = load_pem_private_key(
        Path(f"{fixtures}/root_key.pem").read_bytes(), b"123"
    )
    root_cert = load_pem_x509_certificate(
        Path(f"{fixtures}/root_cert.pem").read_bytes()
    )

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    public_key = private_key.public_key().public_bytes(
        encoding=serialization.Encoding.DER, format=serialization.PublicFormat.PKCS1
    )

    auth_data_public_key = public_key if not wrong_public_key else "XXXXX".encode()

    auth_data = (
        sha256(app_id.encode()).digest()
        + b"\x00"  # Flag, we'll fill it with zero
        + struct.pack("!I", counter)
        + aaguid  # TODO: Fillup to ensure 16 bytes
        + struct.pack("!H", 32)  # Our SHA-256 digest is always 32bytes
        + sha256(auth_data_public_key).digest()
    )

    nonce = sha256(auth_data + sha256(nonce).digest())

    # When comparing this nonce with the one calculated on the server, we'll strip 6 bytes which are normally
    # used to indicated an ASN1 envelope sequence. See the verify_nonce method in the apple verifier.
    der_nonce = bytes(6) + OctetString(nonce.digest()).native

    subject = x509.Name(
        [x509.NameAttribute(NameOID.ORGANIZATION_NAME, "pyattest-testing-leaf")]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(root_cert.subject)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=10))
        .add_extension(key_usage, critical=False)
        .add_extension(
            UnrecognizedExtension(
                ObjectIdentifier("1.2.840.113635.100.8.2"), der_nonce
            ),
            critical=False,
        )
        .sign(root_key, hashes.SHA256())
    )

    data = {
        "fmt": "apple-appattest",
        "attStmt": {
            "x5c": [
                cert.public_bytes(serialization.Encoding.DER),
                root_cert.public_bytes(serialization.Encoding.DER),
            ],
            "receipt": b"",
        },
        "authData": auth_data,
    }

    return cbor_encode(data), public_key



from abc import ABC, abstractmethod


class Config(ABC):
    production = False

    @property
    @abstractmethod
    def attestation_verifier_class(self): ...

    @property
    @abstractmethod
    def assertion_verifier_class(self): ...



import base64

from cryptography.hazmat.primitives.serialization import load_der_public_key
from typing import Optional, List, Set

from pyattest.configs.config import Config
from pyattest.verifiers.google_assertion import GoogleAssertionVerifier
from pyattest.verifiers.google_play_integrity_attestation import (
    GooglePlayIntegrityAttestationVerifier,
)
from pyattest.exceptions import IllegalConfigurationException


class GooglePlayIntegrityApiConfig(Config):
    attestation_verifier_class = GooglePlayIntegrityAttestationVerifier
    assertion_verifier_class = GoogleAssertionVerifier

    def __init__(
        self,
        decryption_key: str,
        verification_key: str,
        apk_package_name: str,
        production: bool,
        allow_non_play_distribution: bool = False,
        verify_code_signature_hex: Optional[List[str]] = None,
        required_device_verdict: str = "MEETS_DEVICE_INTEGRITY",
    ):
        if allow_non_play_distribution and not verify_code_signature_hex:
            raise IllegalConfigurationException(
                "When allowing distribution through channels other than the Play Store you need to "
                + "provide the sha256 digest of your signing certificate! (Obtain via ./gradlew "
                + "signingReport or set production to false for dev builds)"
            )

        self.decryption_key = base64.standard_b64decode(decryption_key)
        self.verification_key = load_der_public_key(
            base64.standard_b64decode(verification_key)
        )
        self.apk_package_name = apk_package_name
        self.production = production
        self.allow_non_play_distribution = allow_non_play_distribution
        if verify_code_signature_hex:
            self.verify_code_signature = [
                self.__convert_signing_digest(hex) for hex in verify_code_signature_hex
            ]
        else:
            self.verify_code_signature = None

        self.required_device_verdict = required_device_verdict

    @staticmethod
    def __convert_signing_digest(hex: str):
        sanitized = hex.replace(":", "")
        digest_bytes = bytearray.fromhex(sanitized)
        base64_signature = base64.urlsafe_b64encode(digest_bytes).decode()
        return base64_signature.replace("=", "")



from abc import ABC, abstractmethod

from pyattest.assertion import Assertion


class AssertionVerifier(ABC):
    def __init__(self, assertion: Assertion):
        self.assertion = assertion

    @abstractmethod
    def verify(self): ...
